/*
** ptsnoop.c
**
** Copyright (c) 2004 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <termios.h>


extern char version[];

#ifndef HAVE_HRTIME_T

#include <sys/time.h>

typedef long long hrtime_t;

hrtime_t
gethrtime(void)
{
    hrtime_t tb;
    struct timeval tvb;

    gettimeofday(&tvb, NULL);
    tb = (long long) tvb.tv_sec * 1000000000 +
	(long long ) tvb.tv_usec * 1000;

    return tb;
}
#endif

char *argv0 = "ptsnoop";

unsigned long bufsize_default = 64*1024;

int print_num = 'x';
int print_char = 0;
int print_bytes_per_line = 16;


unsigned int packet_no = 0;

unsigned int devices = 0;
hrtime_t start_time = 0;

unsigned long blocking_delay = 50000; /* s, default: 0.05s */

struct devtab {
    char *devname;
    int fd;
    int packet;
    hrtime_t start;
    hrtime_t last;
    size_t size;
    size_t length;
    unsigned char *buffer;
} *devtab = NULL;

FILE *output_fp = NULL;


void
usage(FILE *fp)
{
    fprintf(fp, "Usage: %s [options] [<device1> ... <deviceN>]\n", argv0);
    fprintf(fp, "Options:\n");
    fprintf(fp, "\t-h        Usage information\n");
    
    fprintf(fp, "\t-w<num>   Display <num> bytes per line (default: %u)\n",
	    print_bytes_per_line);
							    
    fprintf(fp, "\t-O<chars> Select output options, x=hex, d=dec, o=oct, c=char (Default: x)\n");
							    
    fprintf(fp, "\t-b<num>   Maximum buffer size (default: %lu)\n",
	    (unsigned long) bufsize_default);
    fprintf(fp, "\t-B<num>   Blocking delay in microseconds (default: %lu)\n",
	    (unsigned long) blocking_delay);
							      
    fprintf(fp, "\t-o<file>  Write packets to a special file\n");
    fprintf(fp, "\t-i<file>  Read packets from previously saved file\n");
}



void
print_packet(struct devtab *dtp)
{
    size_t i, j;
    
    
    printf("#%u : Device %s : Packet %u : Length %lu : Start %lu.%09lu\n",
	   packet_no,
	   dtp->devname, dtp->packet,
	   (unsigned long) dtp->length,
	   (unsigned long) (dtp->start / 1000000000),
	   (unsigned long) (dtp->start - (dtp->start / 1000000000) * 1000000000));
    
    i = 0;
    j = 16;
    
    while (i < dtp->length)
    {
	printf("    %5lu", (unsigned long) i);

	if (print_num)
	{
	    printf(" :");
	    
	    for (j = 0; j < print_bytes_per_line; j++)
	    {
		if (i + j < dtp->length)
		{
		    switch (print_num)
		    {
		      case 'x':
			printf(" %02x", dtp->buffer[i+j]);
			break;
			
		      case 'd':
			printf(" %3d", dtp->buffer[i+j]);
			break;
			
		      case 'o':
			printf(" %03o", dtp->buffer[i+j]);
			break;
		    }
		}
		else if (print_char)
		{
		    if (print_num == 'x')
			printf("   ");
		    else
			printf("    ");
		}
	    }
	}

	if (print_char)
	{
	    printf(" : ");

	    for (j = 0; j < print_bytes_per_line && i+j < dtp->length; j++)
	    {
		if (isgraph(dtp->buffer[i+j]) || dtp->buffer[i+j] == ' ')
		    putchar(dtp->buffer[i+j]);
		else
		    putchar('?');
	    }
	}
	
	putchar('\n');

	i += j;

	fflush(stdout);
    }
    
    putchar('\n');
}

void
write_packet(FILE *fp, struct devtab *dtp)
{
    int len;

    len = strlen(dtp->devname);
    if (fwrite(&len, sizeof(len), 1, fp) != 1 ||
	fwrite(dtp->devname, len, 1, fp) != 1 ||
	fwrite(&dtp->packet, sizeof(dtp->packet), 1, fp) != 1 ||
	fwrite(&dtp->start, sizeof(dtp->start), 1, fp) != 1 ||
	fwrite(&dtp->last, sizeof(dtp->last), 1, fp) != 1 ||
	fwrite(&dtp->length, sizeof(dtp->length), 1, fp) != 1 ||
	fwrite(dtp->buffer, dtp->length, 1, fp) != 1)
    {
	fprintf(stderr, "%s: fwrite() failed: %s\n",
		argv0, strerror(errno));
	exit(1);
    }

    fflush(fp);
}


int
read_packet(FILE *fp, struct devtab *dtp)
{
    int len, rc;

    
    if (dtp->devname)
	free(dtp->devname);
    if (dtp->buffer)
	free(dtp->buffer);
    
    rc = fread(&len, sizeof(len), 1, fp);
    if (rc < 1)
	return rc;
    
    dtp->devname = malloc(len+1);
    if (fread(dtp->devname, len, 1, fp) != 1)
	return -1;
    
    dtp->devname[len] = '\0';

    if (fread(&dtp->packet, sizeof(dtp->packet), 1, fp) != 1 || 
	fread(&dtp->start, sizeof(dtp->start), 1, fp) != 1 ||
	fread(&dtp->last, sizeof(dtp->last), 1, fp) != 1 ||
	fread(&dtp->length, sizeof(dtp->length), 1, fp) != 1)
	return -1;

    if (dtp->length > dtp->size)
    {
	dtp->buffer = realloc(dtp->buffer, dtp->length);
	if (dtp->buffer == NULL)
	{
	    fprintf(stderr, "%s: realloc(..., %lu): %s\n",
		    argv0, (unsigned long) dtp->length, strerror(errno));
	    exit(1);
	}
	
	dtp->size = dtp->length;
    }
    
    if (fread(dtp->buffer, dtp->length, 1, fp) != 1)
	return -1;

    return 1;
}


void
handle_packet(struct devtab *dtp)
{
    ++packet_no;
    dtp->start -= start_time;
    dtp->last -= start_time;

    
    if (output_fp)
    {
	write_packet(output_fp, dtp);
	printf("\r%u", packet_no);
	fflush(stdout);
    }
    else
	print_packet(dtp);
}


int
bps2termios(int speed)
{
    switch (speed)
    {
      case 300:
	return B300;
	
      case 600:
	return B600;
	
      case 1200:
	return B1200;
	
      case 2400:
	return B2400;
	
      case 4800:
	return B4800;
	
      case 9600:
	return B9600;
	
      case 19200:
	return B19200;
	
      case 38400:
	return B38400;

      default:
	fprintf(stderr, "%s: invalid bps rate: %u\n", argv0, speed);
	exit(1);
    }
}


int
main(int argc, char *argv[])
{
    int i, j;
    struct termios term; 
    FILE *input_fp = NULL;
    int nextarg = 0;
    char *arg = NULL;
    int bps = 9600;
    

    argv0 = argv[0];
    
    for (i = 1; i < argc && argv[i][0] == '-'; ++i)
    {
	nextarg = 0;
	
	for (j = 1; !nextarg && argv[i][j]; ++j)
	    switch (argv[i][j])
	    {
	      case 'h':
		usage(stdout);
		exit(0);

	      case 'O':
		nextarg = 1;
		
		if (argv[i][j+1])
		    arg = argv[i]+j+1;
		else if (argv[i+1])
		    arg = argv[++i];
		else
		{
		    fprintf(stderr, "%s: Missing argument for -O\n",
			    argv0);
		    exit(1);
		}

		print_num = 0;
		print_char = 0;
		
		for (; *arg; ++arg)
		    switch (*arg)
		    {
		      case 'x':
			print_num = 'x';
			break;
		
		      case 'd':
			print_num = 'd';
			break;
		
		      case 'o':
			print_num = 'o';
			break;
		
		      case 'c':
			print_char++;
			break;
		    }
		break;
		
	      case 'V':
		printf("[ptsnoop, version %s - Copyright (c) 2004 Peter Eriksson <pen@lysator.liu.se>]\n\n", version);
		break;
		
	      case 'S':
		nextarg = 1;
		
		if (argv[i][j+1])
		    arg = argv[i]+j+1;
		else if (argv[i+1])
		    arg = argv[++i];
		else
		{
		    fprintf(stderr, "%s: Missing argument for -S\n",
			    argv0);
		    exit(1);
		}
		
		if (sscanf(arg, "%u", &bps) != 1)
		{
		    fprintf(stderr, "%s: not a number: %s\n",
			    argv0, arg);
		    exit(1);
		}
		break;
		
	      case 'o':
		nextarg = 1;
		
		if (argv[i][j+1])
		    arg = argv[i]+j+1;
		else if (argv[i+1])
		    arg = argv[++i];
		else
		{
		    fprintf(stderr, "%s: Missing file argument for -o\n",
			    argv0);
		    exit(1);
		}
		
		output_fp = fopen(arg, "w");
		if (output_fp == NULL)
		{
		    fprintf(stderr, "%s: fopen: %s: %s\n",
			    argv0, arg, strerror(errno));
		    exit(1);
		}
		break;
		
	      case 'i':
		nextarg = 1;
		
		if (argv[i][j+1])
		    arg = argv[i]+j+1;
		else if (argv[i+1])
		    arg = argv[++i];
		else
		{
		    fprintf(stderr, "%s: Missing file argument for -i\n",
			    argv0);
		    exit(1);
		}
		
		input_fp = fopen(arg, "r");
		if (input_fp == NULL)
		{
		    fprintf(stderr, "%s: fopen: %s: %s\n",
			    argv0, arg, strerror(errno));
		    exit(1);
		}
		break;
		
	      case 'w':
		nextarg = 1;
		
		if (argv[i][j+1])
		    arg = argv[i]+j+1;
		else if (argv[i+1])
		    arg = argv[++i];
		else
		{
		    fprintf(stderr, "%s: Missing argument for -w\n",
			    argv0);
		    exit(1);
		}
		
		if (sscanf(arg, "%u", &print_bytes_per_line) != 1)
		{
		    fprintf(stderr, "%s: not a number: %s\n",
			    argv0, arg);
		    exit(1);
		}
		break;
		
	      case 'b':
		nextarg = 1;
		
		if (argv[i][j+1])
		    arg = argv[i]+j+1;
		else if (argv[i+1])
		    arg = argv[++i];
		else
		{
		    fprintf(stderr, "%s: Missing argument for -b\n",
			    argv0);
		    exit(1);
		}
		
		if (sscanf(arg, "%lu", &bufsize_default) != 1)
		{
		    fprintf(stderr, "%s: not a number: %s\n",
			    argv0, arg);
		    exit(1);
		}
		break;
		
	      case 'B':
		nextarg = 1;
		
		if (argv[i][j+1])
		    arg = argv[i]+j+1;
		else if (argv[i+1])
		    arg = argv[++i];
		else
		{
		    fprintf(stderr, "%s: Missing argument for -B\n",
			    argv0);
		    exit(1);
		}
		
		if (sscanf(arg, "%lu", &blocking_delay) != 1)
		{
		    fprintf(stderr, "%s: not a number: %s\n",
			    argv0, arg);
		    exit(1);
		}
		break;
		
	      default:
		fprintf(stderr, "%s: invalid switch: -%c\n",
			argv0, argv[i][j]);
		exit(1);
	    }
    }
    
    devices = argc - i;

    if (input_fp)
    {
	struct devtab dvb;

	
	memset(&dvb, 0, sizeof(dvb));
	dvb.size = bufsize_default;
	dvb.buffer = malloc(dvb.size);
	
	while (read_packet(input_fp, &dvb) > 0)
	{
	    ++packet_no;
	    
	    if (i < argc)
	    {
		for (j = i; j < argc && strcmp(dvb.devname, argv[j]); j++)
		    ;
		
		if (j == argc)
		    continue;
	    }
	    
	    print_packet(&dvb);
	}
    }
    else
    {
	devtab = calloc(devices, sizeof(devtab[0]));
	if (!devtab)
	{
	    fprintf(stderr, "%s: calloc(%u, %lu): %s\n",
		    argv0, devices, (unsigned long) sizeof(devtab[0]), strerror(errno));
	    exit(1);
	}


	j = 0;
	while (i < argc)
	{
	    devtab[j].devname = strdup(argv[i]);
	    devtab[j].fd = open(argv[i], O_RDONLY|O_NONBLOCK|O_EXCL|O_NOCTTY);
	    if (devtab[j].fd < 0)
	    {
		fprintf(stderr, "%s: open: %s: %s\n",
			argv0, argv[i], strerror(errno));
		exit(1);
	    }

	    if (tcgetattr(devtab[j].fd, &term) < 0)
	    {
		fprintf(stderr, "%s: tcgetattr(%d) [%s]: %s\n",
			argv0, devtab[j].fd, argv[i], strerror(errno));
		exit(1);
	    }
	
	    /*
	     * Set canonical mode and local connection; set specified speed,
	     * 8 bits and no parity; map CR to NL; ignore break.
	     */

	    memset(term.c_cc, 0, 16);
	    term.c_cc[VEOF] = 1;

	    term.c_iflag = IGNBRK | IGNPAR ;
	    term.c_oflag = 0;

	    term.c_cflag = CS8 | bps2termios(bps) | CLOCAL | CREAD;
	    term.c_lflag = IEXTEN;

	    if (tcsetattr(devtab[j].fd, TCSAFLUSH, &term) < 0)
	    {
		fprintf(stderr, "%s: tcsetattr(%d) [%s]: %s\n",
			argv0, devtab[j].fd, argv[i], strerror(errno));
		exit(1);
	    }

	    devtab[j].start = 0;
	    devtab[j].last = 0;
	    devtab[j].length = 0;
	    devtab[j].buffer = malloc(bufsize_default);
	    if (devtab[j].buffer == NULL)
	    {
		fprintf(stderr, "%s: malloc(%lu) [devtab[%u].buffer]: %s\n",
			argv0, (unsigned long) bufsize_default, j, strerror(errno));
		exit(1);
	    }
	    
	    devtab[j].size = bufsize_default;

	    ++j;
	    ++i;
	}


	start_time = gethrtime();

	while (1)
	{
	    int high_fd = 0;
	    int rc = 0;
	    fd_set rfds;
	    struct timeval tvb;
	    hrtime_t now;
	
	
	    FD_ZERO(&rfds);
	    for (j = 0; j < devices; j++)
	    {
		if (devtab[j].fd)
		{
		    ++rc;
		
		    FD_SET(devtab[j].fd, &rfds);
		    if (devtab[j].fd > high_fd)
			high_fd = devtab[j].fd;
		}
	    }

	    if (rc == 0)
		break;

	    tvb.tv_sec = blocking_delay/1000000;
	    tvb.tv_usec = blocking_delay-(tvb.tv_sec*1000000);

	    while ((rc = select(high_fd+1, &rfds, NULL, NULL, &tvb)) < 0 &&
		   errno == EINTR)
		;

	    now = gethrtime();
	
	    for (j = 0; j < devices; j++)
	    {
		if (devtab[j].fd && FD_ISSET(devtab[j].fd, &rfds))
		{
		    /* Data available */

		    if (devtab[j].start)
		    {
			/* Packet in progress */
		    
			devtab[j].last = now;

		    }
		    else
		    {
			/* New packet */
		    
			devtab[j].start = devtab[j].last = now;
		    
		    }
		
		    while ((rc = read(devtab[j].fd,
				      devtab[j].buffer+devtab[j].length,
				      devtab[j].size - devtab[j].length)) < 0 &&
			   errno == EINTR)
			;

		    if (rc < 0)
		    {
			/* Error */

			fprintf(stderr, "%s: read(%u, ..., %lu) [%s]: %s\n",
				argv0,
				devtab[j].fd,
				(unsigned long) (devtab[j].size-devtab[j].length),
				devtab[j].devname,
				strerror(errno));

			if (devtab[j].length > 0)
			    handle_packet(&devtab[j]);
		    
			close(devtab[j].fd);
			devtab[j].fd = 0;
		    }
		    else if (rc == 0)
		    {
			fprintf(stderr, "%s: read(%u, ..., %lu) [%s]: %s\n",
				argv0,
				devtab[j].fd,
				(unsigned long) (devtab[j].size-devtab[j].length),
				devtab[j].devname,
				"EOF");
		    
			if (devtab[j].length > 0)
			    handle_packet(&devtab[j]);
		    
			close(devtab[j].fd);
			devtab[j].fd = 0;
		    }
		    else
		    {
			/* More data */

			devtab[j].length += rc;

			if (devtab[j].length == devtab[j].size)
			{
			    /* Buffer full, print it and start over */
			
			    handle_packet(&devtab[j]);

			    devtab[j].packet++;
			    devtab[j].start = devtab[j].last = 0;
			    devtab[j].length = 0;
			}
		    }
		
		}
		else
		{
		    /* No data available */
		
		    if (devtab[j].start)
		    {
			/* Packet in progress */
		    
			if (now - devtab[j].last >= (blocking_delay*1000))
			{
			    /* End of packet reached */

			    handle_packet(&devtab[j]);
			
			    devtab[j].packet++;
			    devtab[j].start = devtab[j].last = 0;
			    devtab[j].length = 0;
			}
		    }
		}
	    }
	}
    }
    
    return 0;
}

